
#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import json
import logging
from holado_report.report.builders.report_builder import ReportBuilder
import weakref
from holado_python.common.tools.datetime import FORMAT_DATETIME_ISO, DateTime
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_system.system.filesystem.file import File

logger = logging.getLogger(__name__)



class ExecutionHistoricReportBuilder(ReportBuilder):
    def __init__(self, execution_historic, filepath, file_format='json'):
        self.__execution_historic_weakref = weakref.ref(execution_historic)
        self.__file_format = file_format.lower()
        
        if self.__file_format == 'json':
            self.__file = File(filepath, mode='wt')
        else:
            raise TechnicalException(f"Unmanaged format '{self.__file_format}' (possible formats: 'json')")
        
    @property
    def __execution_historic(self):
        return self.__execution_historic_weakref()
    
    def build(self):
        with self.__file as feh:
            eh_json = self.__convert_execution_historic_to_json()
            json_str = json.dumps(eh_json, ensure_ascii=False, indent=4)
            feh.write(json_str)
        
    def __convert_execution_historic_to_json(self):
        return [self.__convert_execution_historic_feature_to_json(eh_feat) for eh_feat in self.__execution_historic]
    
    def __convert_execution_historic_feature_to_json(self, eh_feature):
        return {
            'feature': self.__convert_feature_to_json(eh_feature.feature, eh_feature.feature_context, eh_feature.feature_report),
            'scenarios': [self.__convert_execution_historic_scenario_to_json(eh_sce) for eh_sce in eh_feature.scenarios ]
            }
    
    def __convert_feature_to_json(self, feature, feature_context, feature_report):
        res = {
            'name': feature.name,
            'description': feature.description,
            'tags': [str(tag) for tag in feature.tags],
            'status': feature.status.name,
            'duration': feature.duration,
            'start_date': DateTime.datetime_2_str(feature_context.start_datetime, dt_format=FORMAT_DATETIME_ISO)
            }
        if feature_context.end_datetime:
            res['end_date'] = DateTime.datetime_2_str(feature_context.end_datetime, dt_format=FORMAT_DATETIME_ISO)
        res.update({
            'filename': feature.filename,
            'report': feature_report.report_path
            })
        return res
    
    def __convert_execution_historic_scenario_to_json(self, eh_scenario):
        return {
            'scenario': self.__convert_scenario_to_json(eh_scenario.scenario, eh_scenario.scenario_context, eh_scenario.scenario_report),
            'steps_by_scope': {scope_name:[self.__convert_execution_historic_step_to_json(eh_step) for eh_step in steps] for scope_name, steps in eh_scenario.steps_by_scope.items()},
            'validation_category': eh_scenario.status_info.validation_category if eh_scenario.status_info is not None else None,
            'validation_status': eh_scenario.status_info.validation_status if eh_scenario.status_info is not None else None,
            'step_failed': self.__convert_step_to_json(eh_scenario.status_info.step_failed) if eh_scenario.status_info is not None and eh_scenario.status_info.step_failed is not None else None,
            'step_failed_number': eh_scenario.status_info.step_failed_number if eh_scenario.status_info is not None else None
            }
        
    def __convert_scenario_to_json(self, scenario, scenario_context, scenario_report):
        res = {
            'name': scenario.name,
            'description': scenario.description,
            'tags': [str(tag) for tag in scenario.tags],
            'status': scenario.status.name,
            'duration': scenario.duration,
            'start_date': DateTime.datetime_2_str(scenario_context.start_datetime, dt_format=FORMAT_DATETIME_ISO)
            }
        if scenario_context.end_datetime:
            res['end_date'] = DateTime.datetime_2_str(scenario_context.end_datetime, dt_format=FORMAT_DATETIME_ISO)
        res.update({
            'filename': scenario.filename,
            'line': scenario.line,
            'report': scenario_report.report_path
            })
        return res
        
    def __convert_execution_historic_step_to_json(self, eh_step):
        if eh_step.step is not None:
            res = self.__convert_step_to_json(eh_step.step)
        else:
            res = {}
        if eh_step.step_context is not None:
            if eh_step.step_context.status is not None:
                res['status'] = eh_step.step_context.status
            if eh_step.step_context.start_datetime:
                res['start_date'] = DateTime.datetime_2_str(eh_step.step_context.start_datetime, dt_format=FORMAT_DATETIME_ISO)
            if eh_step.step_context.end_datetime:
                res['end_date'] = DateTime.datetime_2_str(eh_step.step_context.end_datetime, dt_format=FORMAT_DATETIME_ISO)
            
        res['description'] = eh_step.step_description
        if eh_step.sub_steps:
            res['sub_steps'] = [self.__convert_execution_historic_step_to_json(eh_sub_step) for eh_sub_step in eh_step.sub_steps ]
        return res
        
    def __convert_step_to_json(self, step):
        from holado_report.report.report_manager import ReportManager
        res = {
            'status': step.status.name,
            'duration': step.duration,
            }
        step_error = ReportManager.StepTools.get_step_error(step)
        if step_error:
            res['error'] = step_error
        if isinstance(step.filename, str) and step.filename != "<string>":
            res['filename'] = step.filename
            res['line'] = step.line
        return res
        
    
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################



from builtins import range
from builtins import super
from holado_core.common.tools.comparators.comparator import Comparator
from holado_core.common.tools.tools import Tools
from holado_core.common.exceptions.verify_exception import VerifyException
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.exceptions.functional_exception import FunctionalException
import logging
from holado_core.common.exceptions.exceptions import Exceptions

logger = logging.getLogger(__name__)


class TableRowComparator(Comparator):
    def __init__(self, cell_comparator=None, cells_comparators=None):
        """
        Define the method used to compare cells.
        Implementation tests first if cells_comparators is defined, else it uses cell_comparator 
        """
        super().__init__("table row")
        
        self.__cell_comparator = cell_comparator
        self.__cells_comparators = cells_comparators
        if self.__cell_comparator is None and self.__cells_comparators is None:
            raise TechnicalException("No cell comparator is defined")
        
    def equals(self, row_1, row_2, is_obtained_vs_expected=True, raise_exception=True, do_log=True):
        try:
            res = True
            
            # Verify number of cells
            if row_1.nb_cells != row_2.nb_cells:
                if raise_exception:
                    raise VerifyException("Rows have not same number of cells (row {name1}: {nb1} cells ; row {name2}: {nb2} cells):\n  row {name1}:\n{repr1}\n  row {name2}:\n{repr2})".format(
                                                   name1=self._get_name_1(is_obtained_vs_expected), nb1=row_1.nb_cells, repr1=row_1.represent(0), 
                                                   name2=self._get_name_2(is_obtained_vs_expected), nb2=row_2.nb_cells, repr2=row_2.represent(0)))
                else:
                    return False
            if self.__cells_comparators is not None and row_1.nb_cells > len(self.__cells_comparators):
                raise TechnicalException("Rows have more cells than defined in list of cells comparators")
                
                
            # Compare cells content
            for i in range(row_1.nb_cells):
                # Define cell comparator to use
                if self.__cells_comparators is not None:
                    cell_comp = self.__cells_comparators[i]
                else:
                    cell_comp = self.__cell_comparator
                
                # Manage case where the cell comparator is None, meaning the cells mustn't be compared
                if cell_comp is None:
                    continue
                
                # Compare cells
                try:
                    res = cell_comp.equals(row_1.get_cell(i),
                                           row_2.get_cell(i),
                                           is_obtained_vs_expected = is_obtained_vs_expected,
                                           raise_exception = raise_exception,
                                           do_log=do_log)
                except FunctionalException as exc:
                    msg = "Difference exists in row at cells of index {}:\n{}\n  row {}: {}\n  row {}: {}\n  cell comparator: {}".format(
                            i, Tools.indent_string(4, Exceptions.exception_message(exc)), 
                            self._get_name_1(is_obtained_vs_expected), self.__represent_row(row_1), 
                            self._get_name_2(is_obtained_vs_expected), self.__represent_row(row_2),
                            cell_comp)
                    raise VerifyException(msg) from exc
                except Exception as exc:
                    msg = "Error while comparing cells of index {}:\n{}\n  row {}: {}\n  row {}: {}\n  cell comparator: {}".format(
                            i, Tools.indent_string(4, Exceptions.exception_message(exc)), 
                            self._get_name_1(is_obtained_vs_expected), self.__represent_row(row_1), 
                            self._get_name_2(is_obtained_vs_expected), self.__represent_row(row_2),
                            cell_comp)
                    raise TechnicalException(msg) from exc
                
                if not res:
                    if raise_exception:
                        raise TechnicalException("This case should have already raised an exception")
                    else:
                        if do_log and Tools.do_log(logger, logging.DEBUG):
                            logger.debug("Difference exists in row at cells of index {}:\n  row {}:\n{}\n  row {}:\n{}".format(
                                         i, self._get_name_1(is_obtained_vs_expected), self.__represent_row(row_1, 4), self._get_name_2(is_obtained_vs_expected), self.__represent_row(row_2, 4)))
                        break
                    
            return res
        except (FunctionalException, TechnicalException) as exc:
            raise exc
        except Exception as exc:
            msg = "Error while comparing rows:\n{}\n  row {}: {}\n  row {}: {}".format(
                    Tools.indent_string(4, Exceptions.exception_message(exc)), 
                    self._get_name_1(is_obtained_vs_expected), self.__represent_row(row_1), 
                    self._get_name_2(is_obtained_vs_expected), self.__represent_row(row_2))
            raise TechnicalException(msg) from exc
    
    def __represent_row(self, row, indent=0):
        try:
            return row.represent(indent)
        except Exception as exc:
            return f"[ERROR while representing row: {Exceptions.exception_message(exc)}]"
        





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import json
from holado_core.common.exceptions.technical_exception import TechnicalException
from json.decoder import JSONDecodeError
import base64
from holado_python.standard_library.typing import Typing
from typing import get_args, get_type_hints
from holado_core.common.handlers.features.resource_by_type import FeatureClassResourceByType
from holado.common.handlers.undefined import undefined_value
import inspect

logger = logging.getLogger(__name__)


# Define managed json types

class JsonTypes(FeatureClassResourceByType):
    @classmethod
    def _get_class_resource_description(cls, plural=False):
        if plural:
            return 'json types'
        else:
            return 'json type'
    
    @classmethod
    def _new_parents(cls, parents, class_, **kwargs):
        if parents is None:
            res = []
        else:
            res = list(parents)
        res.append( (class_, kwargs) )
        return res


# Register default json types

JsonTypes.register_resource_for_type_in_class('primitive', lambda c: c in [int, float, bool, type(None)], None, 
                   (lambda x: x, lambda x: x, None))
JsonTypes.register_resource_for_type_in_class('bytes', lambda c: inspect.isclass(c) and issubclass(c, bytes), None, 
                   (base64.b64decode, base64.b64encode, None))


def _to_json_from_str(value):
    try:
        return json.loads(value)
    except JSONDecodeError:
        return value

JsonTypes.register_resource_for_type_in_class('str', lambda c: inspect.isclass(c) and issubclass(c, str), None, 
                   (lambda x: x, _to_json_from_str, None))


def _from_json_to_Dict(obj_json, class_, _parents=None, converter=None):
    if not isinstance(obj_json, dict):
        raise TechnicalException(f"Argument 'obj_json' must be a dict")
    if converter is None:
        from holado_json.ipc.json_converter import JsonConverter
        converter = JsonConverter()
    
    # Define key and value types if they are available, else type of of key and value will be used
    args = get_args(class_)
    if len(args) > 0:
        if len(args) != 2:
            raise TechnicalException(f"Unmanaged convert from json to {Typing.get_class_fullname(class_)}: 2 type arguments are expected (obtained type arguments: {args})")
        key_type, value_type = args
    else:
        # Use object type of each dict key & value
        key_type, value_type = None, None
    
    res = {}
    for k, v in obj_json.items():
        key = converter.from_json_to(k, key_type, _parents=JsonTypes._new_parents(_parents, class_, name=k, index=0))
        value = converter.from_json_to(v, value_type, _parents=JsonTypes._new_parents(_parents, class_, name=k, index=1))
        res[key] = value
    
    return res

def _to_json_from_dict_or_list(value):
    try:
        tmp = json.dumps(value)
        return json.loads(tmp)
    except JSONDecodeError:
        return value
    
JsonTypes.register_resource_for_type_in_class('Dict', lambda c: inspect.isclass(c) and issubclass(c, dict), None, 
                   (None, _to_json_from_dict_or_list, _from_json_to_Dict))


def _from_json_to_List(obj_json, class_, _parents=None, converter=None):
    if not isinstance(obj_json, list):
        raise TechnicalException(f"Argument 'obj_json' must be a list")
    if converter is None:
        from holado_json.ipc.json_converter import JsonConverter
        converter = JsonConverter()
    
    element_type = get_args(class_)[0]
    
    res = []
    for index, el in enumerate(obj_json):
        res_el = converter.from_json_to(el, element_type, _parents=JsonTypes._new_parents(_parents, class_, index=index))
        res.append(res_el)
    
    return res

JsonTypes.register_resource_for_type_in_class('List', lambda c: inspect.isclass(c) and issubclass(c, list), None, 
                        (None, _to_json_from_dict_or_list, _from_json_to_List))


def _from_json_to_NamedTuple(obj_json, class_, _parents=None, converter=None):
    if not isinstance(obj_json, dict):
        raise TechnicalException(f"Argument 'obj_json' must be a dict")
    if converter is None:
        from holado_json.ipc.json_converter import JsonConverter
        converter = JsonConverter()
    
    kwargs = {}
    for name, type_ in get_type_hints(class_).items():
        if name not in obj_json:
            raise TechnicalException(f"No field '{name}' in json object (existing fields: {list(obj_json.keys())})")
        value = converter.from_json_to(obj_json[name], type_, _parents=JsonTypes._new_parents(_parents, class_, name=name))
        kwargs[name] = value
        
    return class_(**kwargs)

JsonTypes.register_resource_for_type_in_class('NamedTuple', lambda c: Typing.is_NamedTuple(class_=c), None, 
                        (None, None, _from_json_to_NamedTuple))


def _from_json_to_object(obj_json, class_, _parents=None, converter=None):
    if not isinstance(obj_json, dict):
        raise TechnicalException(f"Argument 'obj_json' must be a dict")
    if converter is None:
        from holado_json.ipc.json_converter import JsonConverter
        converter = JsonConverter()
    
    res = undefined_value
    
    # Try with constructor having all dict keys as arguments
    try:
        res = class_(**obj_json)
    except TypeError as exc:  # @UnusedVariable
        # logger.print(f"++++++++++ class_(**obj_json): {exc}")
        pass
    
    if res is undefined_value:
        try:
            res = class_()
        except TypeError as exc:  # @UnusedVariable
            # logger.print(f"++++++++++ class_(): {exc}")
            pass
        if res is not undefined_value:
            for k, v in obj_json.items():
                if not isinstance(k, str):
                    raise TechnicalException(f"Failed to set a field in object of type {class_}: a field has name [{k}] that is not a string (from json object: {obj_json})")
                value = converter.from_json_to(v, None, _parents=JsonTypes._new_parents(_parents, class_, name=k, index=1))
                if hasattr(res, k):
                    setattr(res, k, value)
                else:
                    raise TechnicalException(f"Failed to set field '{k}' in object of type {class_}: attribute '{k}' doesn't exists")
    
    if res is undefined_value:
        raise TechnicalException(f"Failed to create {class_} instance from json object {obj_json}")
    return res

JsonTypes.register_resource_for_type_in_class('object', lambda c: inspect.isclass(c) and issubclass(c, object), lambda o: isinstance(o, object) and hasattr(o, '__dict__'), 
                        (None, lambda x: x.__dict__, _from_json_to_object))
    



